/** @preserve   Eloquence WEBDLG v2
   (C) Copyright Marxmeier Software AG, 2020-2021
   Version: 1.925 2021-03-16
   $Id: 00-base.js,v 28.21 2021/03/16 16:15:45 mike Exp $
*//*======================================================================
   WEBDLG v2 namespace / globals
========================================================================*/

var eq = {
   c : {},           // Classes
   p : {             // Properties:
      version : 1925 //  Version 1.925 (major*1,000 + minor)
   }
};

/*========================================================================
   class Obj
   Generic base class.
========================================================================*/

eq.c.Obj = function() {};

/*------------------------------------------------------------------------
   static Obj.subClass()
   Subclassing convenience helper.
------------------------------------------------------------------------*/

(eq.c.Obj.setClassName = function(name, cls) {
   var pd = Object.getOwnPropertyDescriptor(cls.prototype.constructor, 'name');
   if (pd !== undefined)
      pd.value = name;
   else
      pd = {
         value : name,
         configurable : true,
         enumerable : false,
         writable : false
      };
   Object.defineProperty(cls.prototype.constructor, 'name', pd);
})("Obj", eq.c.Obj);

eq.c.Obj.subClass = function(name, cls) {
   var base = this;
   cls.prototype = Object.create(base.prototype);
   Object.defineProperty(cls.prototype, 'constructor', {
      value : cls,
      configurable : false,
      enumerable : false,
      writable : true
   });
   eq.c.Obj.setClassName(name, cls);
   cls.subClass = function(n, c) { return base.subClass.call(this, n, c) };
   return eq.c[name] = cls;
};

/** @preserve $Id: 05-const.js,v 28.14 2021/03/16 16:16:40 mike Exp $
*//*======================================================================
   static MainOp
   API->Main message opcodes
========================================================================*/

eq.MainOp = {
   applicationStopped : 1,  // Data: restartable true/false
   applicationFailed  : 2,  // Data: none
   cancelCurrentCall  : 3,  // Data: none
   resume             : 4,  // Data: none
   addControl         : 5,  // Data: dialog id, DOM element template
   modifyControl      : 6,  // Data: dialog id, DOM element template
   deleteControl      : 7,  // Data: dialog id, element id
   invalidate         : 8,  // Data: dialog id array
   dlg_DRAW           : 9,  // Data: dialog id, focus element id
   dlg_DO             : 10, // Data: dialog id
   dlg_POPUP_BOX      : 11  // Data: POPUP BOX descriptor
};

/*========================================================================
   static ApiOp
   Main->API message opcodes
========================================================================*/

eq.ApiOp = {
   documentComplete : 1, // Data: layout grid
   openWebSocket    : 2, // Data: WebSocket URL
   onEvent          : 3, // Data: element id, event type
   onPopupEvent     : 4, // Data: button number
   invalidate       : 5  // Data: dialog id
};

/*========================================================================
   static DlgRq.tag
   DLG request tags.
========================================================================*/

eq.RqTag  = {
   child : 6,  // Child control

   // Common attributes.

   cssid      : 7,  // CSS id
   cssclsDel  : 8,  // Delete CSS class(es)
   cssclsAdd  : 9,  // Add CSS class(es)
   posSize    : 10, // Position/size
   rule       : 11, // Rule
   tabonenter : 12, // Enter key navigates focus
   visible    : 13, // Visibility
   sensitive  : 14, // Sensitivity
   tabStop    : 15, // Tab stop
   bgColor    : 16, // Background color
   fgColor    : 17, // Foreground color
   url        : 18, // Url

   // Dialog attributes.

   // Button, StaticText attributes.

   text       : 40, // Text
   textPos    : 41, // Text position
   align      : 42, // Align
   active     : 43, // Active
   radioGroup : 44, // Radio button group

   // EditText attributes.

   multiline  : 50, // Multiline
   readonly   : 51, // Read-only
   content    : 52, // Content
   ispassword : 53, // IsPassword

   // ListBox attributes.

   activeLine : 60, // Active line
   topItem    : 61  // First visible line
};

/*========================================================================
   static RsValType
   Response changed element value type.
========================================================================*/

eq.RsValType = {
   idFocus       : 1, // Focus object id, STRING
   textValue     : 2, // Text value, STRING
   buttonChecked : 3, // Button checked, UINT
   caretPos      : 4, // Caret position, UINT
   activeLine    : 5, // Active line, UINT
   topItem       : 6  // First visible line, UINT
};

/** @preserve $Id: 90-popup.js,v 28.9 2021/03/14 13:39:48 mike Exp $
*//*======================================================================
   class PopupBox extends Obj
   POPUP BOX class.
========================================================================*/

eq.c.Obj.subClass('PopupBox', function(pb, parent) {
   eq.c.Obj.call(this);

   if (eq.c.PopupBox.singleton)
      throw new Error("BUG: POPUP BOX singleton exists");

   if (!parent.classList.contains('eq-idle'))
      throw new Error("BUG: POPUP BOX parent must be idle pane");

   this.popup = document.createElement('DIV');
   this.popup.className =
      'eq-popup-box eq-nselect eq-box-border eq-box-shadow eq-corners';
   this.popup.style.zIndex = eq.Dom.zIndex(parent) + 1;

   // Title.
   var el = document.createElement('SPAN');
   el.className = 'eq-caption';
   if (pb.title)
      eq.Dom.setContent(el, pb.title);
   else
      el.textContent = "Eloquence";
   el.addEventListener('mousedown', eq.c.PopupBox.beginMove);
   this.popup.appendChild(el);

   // Message.
   var msg = pb.msg.split("\n");
   for (var i = 0, l = msg.length; i < l; i++) {
      el = document.createElement('P');
      eq.Dom.setContent(el, msg[i]);
      this.popup.appendChild(el);
   }

   // Buttons.
   var
      btn = document.createElement('DIV'),
      btnDefault = pb.btnDefault;
   btn.className = 'eq-pushbutton';
   btnDefault = btnDefault > 0 ? btnDefault - 1 : 0;
   for (var i = 0, l = pb.btn.length; i < l; i++) {
      el = document.createElement('BUTTON');
      el.type = 'button';
      el.className = 'eq-corners';
      eq.Dom.setContent(el, pb.btn[i]);
      if (i === btnDefault)
         el.classList.add('eq-focused');
      btn.appendChild(el);
   }
   this.popup.appendChild(btn);

   // Add to DOM, top-level, invisible.
   this.popup.style.visibility = 'hidden';
   document.body.appendChild(this.popup);

   // Initialize position, centered to parent if possible.
   var
      rcPopup = this.popup.getBoundingClientRect(),
      rcParent = parent.getBoundingClientRect(),
      rcView = eq.Dom.viewRect(),
      hMax = rcView.h * .9;
   if (rcPopup.height > hMax)
   {
      // Too tall, remnove message line(s).
      el = this.popup.getElementsByTagName('P');
      while (rcPopup.height > hMax && el.length > 0) {
         this.popup.removeChild(el[el.length - 1]);
         rcPopup = this.popup.getBoundingClientRect();
      }
      if (el.length > 0) {
         el[el.length - 1].textContent = "...";
         rcPopup = this.popup.getBoundingClientRect();
      }
   }
   this.rc = {
      x : (rcParent.width - rcPopup.width) / 2 + rcView.x + rcParent.left,
      y : (rcParent.height - rcPopup.height) / 2 + rcView.y + rcParent.top,
      w : rcPopup.width,
      h : rcPopup.height
   };
   this.setPos(rcView);

   // Assign singleton, make visible, set focus to default button.
   eq.c.PopupBox.singleton = this;
   this.popup.style.visibility = '';
   this.popup.querySelector(
      '.eq-popup-box>.eq-pushbutton>.eq-focused').focus();
});

/*------------------------------------------------------------------------
   static PopupBox.singleton
   Global POPUP BOX instance.
------------------------------------------------------------------------*/

eq.c.PopupBox.singleton = null;

/*------------------------------------------------------------------------
   PopupBox.close()
   Close POPUP BOX, submit event.
------------------------------------------------------------------------*/

eq.c.PopupBox.prototype.close = function(e) {
   if (e !== undefined) {
      // Click event?
      if (e.type !== 'click')
         return false;

      var i = this.buttonIndex(e.target);
      if (i === -1) {
         // Not a button.
         return false;
      }

      // Notify API thread.
      eq.app.api.postMessage({
         o : eq.ApiOp.onPopupEvent,
         d : i + 1
      });

      e.stopPropagation();
   }

   // Remove from DOM.
   this.popup.querySelector('.eq-popup-box>.eq-caption')
      .removeEventListener('mousedown', eq.c.PopupBox.beginMove);
   this.popup.parentNode.removeChild(this.popup);

   // Remove toplevel idle pane if present.
   var el = document.querySelector('body>.eq-idle');
   if (el)
      el.parentNode.removeChild(el);

   // Clear singleton and clean up.
   eq.c.PopupBox.singleton = null;
   delete this.popup;
   delete this.rc;
   delete this.move;
   return true;
};

/*------------------------------------------------------------------------
   PopupBox.setPos()
   Set POPUP BOX position.
------------------------------------------------------------------------*/

eq.c.PopupBox.prototype.setPos = function(view) {
   view = view || eq.Dom.viewRect();
   var x = this.rc.x, y = this.rc.y;
   if (x + this.rc.w >= view.x + view.w)
      x = view.x + view.w - this.rc.w - 1;
   if (x <= view.x)
      x = view.x + 1;

   if (y + this.rc.h >= view.y + view.h)
      y = view.y + view.h - this.rc.h - 1;
   if (y <= view.y)
      y = view.y + 1;

   this.popup.style.top = y + 'px';
   this.popup.style.left = x + 'px';
};

/*------------------------------------------------------------------------
   static PopupBox.beginMove() .endMove() .onMove()
   Move POPUP BOX.
------------------------------------------------------------------------*/

eq.c.PopupBox.beginMove = function(e) {
   e.stopPropagation();
   eq.c.PopupBox.singleton.move = {
      x : e.clientX,
      y : e.clientY
   };
   addEventListener('mousemove', eq.c.PopupBox.onMove);
   addEventListener('mouseup', eq.c.PopupBox.endMove);
};

eq.c.PopupBox.endMove = function(e) {
   e.stopPropagation();
   removeEventListener('mousemove', eq.c.PopupBox.onMove);
   removeEventListener('mouseup', eq.c.PopupBox.endMove);
   delete eq.c.PopupBox.singleton.move;
};

eq.c.PopupBox.onMove = function(e) {
   e.stopPropagation();

   var self = eq.c.PopupBox.singleton;
   self.rc.x += e.clientX - self.move.x;
   self.rc.y += e.clientY - self.move.y;

   self.setPos();

   self.move.x = e.clientX;
   self.move.y = e.clientY;
};

/*------------------------------------------------------------------------
   PopupBox.onKey()
   POPUP BOX keyboard event handler.
------------------------------------------------------------------------*/

eq.c.PopupBox.prototype.onKey = function(e) {
   var t = e.target;
   if (   this.isEvent(e)
       || t.classList.contains('eq-idle')
       || t === document.body) {
      var i = 0;
      switch (e.key) {
         case 'Enter':
         case ' ':
         case 'Spacebar': // IE, Firefox <= 37
            if (this.buttonIndex(t) !== -1)
               eq.Dom.fireClick(t);
            break;
         case 'Tab':
            i = e.shiftKey ? -1 : 1;
            break;
         case 'ArrowUp':
         case 'Up':
         case 'ArrowLeft':
         case 'Left':
            i = -1;
            break;
         case 'ArrowDown':
         case 'Down':
         case 'ArrowRight':
         case 'Right':
            i = 1;
            break;
         default:
            return false;
      }

      if (i !== 0) {
         var
            btn = this.buttonIndex(t),
            el = this.popup.querySelector(
               'body>.eq-popup-box>.eq-pushbutton').childNodes,
            l = el.length;
         if ((btn += i) < 0)
            btn = l - 1;
         else if (btn >= l)
            btn = 0;
         for (i = 0; i < l; i++)
            el[i].classList.remove('eq-focused');
         el[btn].classList.add('eq-focused');
         el[btn].focus();
      }

      eq.c.App.debounceKey(e.key);
      return true; // Consume event.
   }

   return false;
};

/*------------------------------------------------------------------------
   PopupBox.isEvent()
   Check if event applies to POPUP BOX.
------------------------------------------------------------------------*/

eq.c.PopupBox.prototype.isEvent = function(e) {
   var el = e.target;
   while (el && el !== this.popup)
      el = el.parentNode;
   return !!el;
};

/*------------------------------------------------------------------------
   PopupBox.buttonIndex()
   Get POPUP BOX button index of specified target.
------------------------------------------------------------------------*/

eq.c.PopupBox.prototype.buttonIndex = function(t) {
   var el = this.popup.querySelector(
      'body>.eq-popup-box>.eq-pushbutton').childNodes;
   for (var i = 0, l = el.length; i < l; i++)
      if (t === el[i])
         return i;
   return -1;
};

/** @preserve $Id: 95-util.js,v 28.11 2021/03/16 16:19:57 mike Exp $
*//*======================================================================
   static Dom
   DOM utilities.
========================================================================*/

eq.Dom = {};

/*------------------------------------------------------------------------
   static Dom.viewRect()
   Obtain current view rectangle.
------------------------------------------------------------------------*/

eq.Dom.viewRect = function() {
   return {
      x : pageXOffset,
      y : pageYOffset,
      w : document.documentElement.clientWidth,
      h : document.documentElement.clientHeight
   };
};

/*------------------------------------------------------------------------
   static Dom.setContent()
   Set element text or HTML content.
------------------------------------------------------------------------*/

eq.Dom.setContent = function(el, cn) {
   var self = this.setContent, html = self.isHtml.exec(cn);
   if (html) {
      var
         body = self.parseHtml.parseFromString(html[1], 'text/html').body,
         child;
      while (child = el.lastChild)
         el.removeChild(child);
      while (child = body.firstChild)
         el.appendChild(child);
   }
   else
      el.textContent = cn;
};

eq.Dom.setContent.isHtml = new RegExp('^<html>(.*)(</html>)?$', 'i');
eq.Dom.setContent.parseHtml = new DOMParser();

/*------------------------------------------------------------------------
   static Dom.zIndex()
   Return numeric element z-index.
------------------------------------------------------------------------*/

eq.Dom.zIndex = function(el) {
   var z = getComputedStyle(el).zIndex;
   return +z === +z ? +z : 0;
};

/*------------------------------------------------------------------------
   static Dom.eqControl()
   Obtain 'eq-control' element.
------------------------------------------------------------------------*/

eq.Dom.eqControl = function(el) {
   var c;
   while (el && el !== document && (c = el.classList) !== undefined) {
      if (c.contains('eq-control'))
         return el;
      if (c.contains('eq-container'))
         break;
      el = el.parentNode;
   }
   return null;
}

/*------------------------------------------------------------------------
   static Dom.consumeEvent()
   Consume event.
------------------------------------------------------------------------*/

eq.Dom.consumeEvent = function(e) {
   e.preventDefault();
   e.stopPropagation();
};

/*------------------------------------------------------------------------
   static Dom.fireMouseEvent()
   Fire mouse event on element.
------------------------------------------------------------------------*/

eq.Dom.fireMouseEvent = function(t, ty) {
   var e = document.createEvent("MouseEvent");
   e.initEvent(ty, true, true);
   t.dispatchEvent(e);
};

/*------------------------------------------------------------------------
   static Dom.fireClick() .fireDoubleClick()
   Fire mouse click/double-click on element.
------------------------------------------------------------------------*/

eq.Dom.fireClick = function(t) {
   this.fireMouseEvent(t, 'mousedown');
   this.fireMouseEvent(t, 'mouseup');
   this.fireMouseEvent(t, 'click');
};

eq.Dom.fireDoubleClick = function(t) {
   this.fireMouseEvent(t, 'mousedown');
   this.fireMouseEvent(t, 'mouseup');
   this.fireMouseEvent(t, 'dblclick');
};

/*------------------------------------------------------------------------
   static Dom.isEnterOrSpace()
   Check for ENTER or SPACE key.
------------------------------------------------------------------------*/

eq.Dom.isEnterOrSpace = function(e) {
   switch (e.key) {
      case 'Enter':
      case ' ':
      case 'Spacebar': // IE, Firefox <= 37
         return true;
   }
   return false;
};

/*------------------------------------------------------------------------
   static Dom.onNextCycle()
   Invoke callback on next event cycle.
------------------------------------------------------------------------*/

eq.Dom.onNextCycle = function(cb, arg) {
   var q = this.nextCycleQueue;
   if (q.length === 0)
      setTimeout(this.nextCycleExec);
   q.push({ cb : cb, arg : arg });
};

eq.Dom.nextCycleQueue = [];

eq.Dom.nextCycleExec = function() {
   var q = eq.Dom.nextCycleQueue;
   for (var i = 0, l = q.length; i < l; i++) {
      var c = q[i], a = c.arg;
      if (a !== undefined)
         c.cb(a);
      else
         c.cb();
   }
   q.length = 0;
};

/*========================================================================
   DLG element callbacks.
========================================================================*/

eq.setAttr = new Map();
eq.Focus = new Map();
eq.OnKey = new Map();
eq.OnChange = new Map();
eq.Changed = new Map();

/*------------------------------------------------------------------------
   CheckBox callbacks.
------------------------------------------------------------------------*/

eq.Focus.set('CheckBox', function(el) {
   var n = el.getElementsByTagName('INPUT');
   if (n.length !== 1)
      throw new Error("BUG: Focus.CheckBox id:" + el.id +
         " elements:" + n.length + ", expected: 1");
   return n[0];
});

eq.OnKey.set('CheckBox', function(id, el, e) {
   var dm = eq.Dom;
   if (dm.isEnterOrSpace(e)) {
      dm.fireClick(e.target);
      eq.c.App.debounceKey(e.key);
      return true; // Consume event.
   }
   return false;
});

eq.Changed.set('CheckBox', function(id, el) {
   if (el.length !== 1)
      throw new Error("BUG: Changed.CheckBox id:" + id +
         " elements:" + el.length + ", expected: 1");
   var ch = {
      id : id,
      ty : eq.RsValType.buttonChecked,
      iv : el[0].checked ? 1 : 0
   };
   return ch;
});

/*------------------------------------------------------------------------
   EditText callbacks.
------------------------------------------------------------------------*/

eq.Focus.set('EditText', function(el) {
   var n = el.getElementsByTagName('INPUT');
   if (n.length === 0)
      n = el.getElementsByTagName('TEXTAREA');
   if (n.length !== 1)
      throw new Error("BUG: Focus.EditText id:" + el.id +
         " elements:" + n.length + ", expected: 1");
   return n[0];
});

eq.Changed.set('EditText', function(id, el, ty) {
   if (el.length !== 1)
      throw new Error("BUG: Changed.EditText id:" + id +
         " elements:" + el.length + ", expected: 1");
   var v = [];
   for (var i = 0, l = ty.length; i < l; i++)
      switch (ty[i]) {
         case 'input':
            v.push({
               ty : eq.RsValType.textValue,
               sv : el[0].value
            });
            break;

         case 'blur': {
            var iv = el[0].selectionStart;
            if (iv >= 0)
               v.push({
                  ty : eq.RsValType.caretPos,
                  iv : iv
               });
            break;
         }
      }
   switch (v.length) {
      case 0:
         return null;
      case 1:
         v[0].id = id;
         return v[0];
      default:
         return {
            id : id,
            v : v
         };
   }
});

/*------------------------------------------------------------------------
   ListBox callbacks.
------------------------------------------------------------------------*/

eq.Dom.ListBoxSelect = function(el, ln) {
   /* deselect */
   if (el = el.querySelector('.eq-column>.eq-selected'))
      el.classList.remove('eq-selected');
   /* select */
   if (   ln !== undefined
       && (el = ln.parentNode)
       && el.classList.contains('eq-column')) {
      ln.classList.add('eq-selected');
   }
};

eq.Dom.ListBoxSet = function(rt) {
   var tag = eq.RqTag;
   switch (rt.t) {
      case tag.activeLine: {
         var dm = eq.Dom, el = rt.el, v = rt.v;
         if (v > 0) {
            var n = el.querySelector('.eq-listbox>.eq-column').childNodes;
            if (v <= n.length) {
               dm.ListBoxSelect(el, n[v - 1]);
               break;
            }
         }
         dm.ListBoxSelect(el);
         break;
      }

      case tag.topItem: {
         var
            el = rt.el,
            n = el.querySelector('.eq-listbox>.eq-column').childNodes,
            c, bx, mh, lh, st = 0;
         if (n.length > 1) {
            c = n[0].parentNode;
            mh = c.clientHeight - el.clientHeight;
            lh = n[1].offsetTop - n[0].offsetTop;
            if (mh && lh && (st = rt.v * lh) > mh)
               st = mh;
         }
         el.scrollTop = st;
         break;
      }

      default:
         throw new Error("ListBoxSet: invalid tag " + rt.t);
   }
};

eq.setAttr.set('ListBox', function(el, rt) {
   var dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.activeLine:
         case tag.topItem:
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(dm.ListBoxSet, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;

         default:
            throw new Error("setAttr.ListBox: invalid tag " + rc.t);
      }
   }
});

eq.OnKey.set('ListBox', function(id, el, e) {
   // Click?
   var dm = eq.Dom;
   if (dm.isEnterOrSpace(e)) {
      dm.fireClick(e.target);
      eq.c.App.debounceKey(e.key);
      return true; // Consume event.
   }

   // Navigation?
   switch (e.key) {
      case 'Home':
      case 'End':
      case 'PageUp':
      case 'PageDown':
      case 'ArrowUp':
      case 'Up':
      case 'ArrowDown':
      case 'Down':
         break;
      default:
         // No.
         return false;
   }

   var
      n = el.querySelector('.eq-listbox>.eq-column').childNodes,
      d = 0,
      p = -1;
   if (n.length) {
      switch (e.key) {
         case 'Home':
            p = 0;
            break;
         case 'End':
            p = n.length - 1;
            break;
         case 'PageUp':
            d = -10;
            break;
         case 'PageDown':
            d = 10;
            break;
         case 'ArrowUp':
         case 'Up':
            d = -1;
            break;
         case 'ArrowDown':
         case 'Down':
            d = 1;
            break;
      }
      var ln = null, bx, bh, lh;
      if (p !== -1)
         ln = n[p];
      else if (d !== 0) {
         if ((d < -1 || d > 1) && n.length > 1) {
            // Page up/down, more than one line,
            // calculate number of visible lines.
            bx = n[0].parentNode.parentNode;
            bh = bx.clientHeight;
            lh = n[1].offsetTop - n[0].offsetTop;
            if (bh && lh)
               d = d > 0 ? Math.ceil(bh / lh) : Math.floor(-(bh / lh));
         }
         var s = el.querySelector('.eq-listbox>.eq-column>.eq-selected');
         if (s) {
            p = 0;
            while (s = s.previousSibling)
               p++;

            if ((p += d) < 0)
               p = 0;
            else if (p >= n.length)
               p = n.length - 1;

            ln = n[p];
         }
         else if (d > 0)
            ln = n[0];
      }

      // Select line if necessary.
      if (ln !== null && !ln.classList.contains('eq-selected')) {
         eq.app.setChanged(id, 'mousedown');
         dm.ListBoxSelect(el, ln);
         if (n.length > 1) {
            // Scroll into view if necessary.
            if (bx === undefined) {
               bx = n[0].parentNode.parentNode;
               bh = bx.clientHeight;
               lh = n[1].offsetTop - n[0].offsetTop;
            }
            if (bh && lh) {
               var bt = bx.scrollTop, lt = ln.offsetTop;
               if (lt < bt)
                  bx.scrollTop = lt;
               else if (lt + lh > bt + bh)
                  bx.scrollTop = lt + lh - bh;
            }
         }
      }
   }

   eq.app.lastKey = e.key;
   return true; // Consume event.
});

eq.OnChange.set('ListBox', function(id, el, e) {
   if (e.type === 'mousedown') {
      var t = e.target, p;
      if (t && !t.classList.contains('eq-listbox'))
         eq.Dom.ListBoxSelect(el, t);
   }
});

eq.Changed.set('ListBox', function(id, el, ty) {
   if (el.length !== 1)
      throw new Error("BUG: Changed.ListBox id:" + id +
         " elements:" + el.length + ", expected: 1");
   var v = [], activeLine = false, topItem = false;
   for (var i = 0, l = ty.length; i < l; i++)
      switch (ty[i]) {
         case 'keydown':
         case 'mousedown':
            if (!activeLine) {
               activeLine = true;
               var
                  ln,
                  ch = {
                     ty : eq.RsValType.activeLine
                  };
               if (ln = el[0].querySelector('.eq-selected')) {
                  var i = 1;
                  while (ln = ln.previousSibling)
                     i++;
                  ch.iv = i;
               }
               else
                  ch.iv = 0;
               v.push(ch);
            }
            // FALLTHROUGH

         case 'blur':
            if (!topItem) {
               topItem = true;
               var
                  n = el[0].childNodes,
                  bx, lh,
                  ch = {
                     ty : eq.RsValType.topItem,
                     iv : 0
                  };
               if (n.length > 1) {
                  bx = n[0].parentNode.parentNode;
                  lh = n[1].offsetTop - n[0].offsetTop;
                  if (lh)
                     ch.iv = Math.ceil(bx.scrollTop / lh);
               }
               v.push(ch);
            }
            break;
      }
   switch (v.length) {
      case 0:
         return null;
      case 1:
         v[0].id = id;
         return v[0];
      default:
         return {
            id : id,
            v : v
         };
   }
});

/*------------------------------------------------------------------------
   PushButton callbacks.
------------------------------------------------------------------------*/

eq.Focus.set('PushButton', function(el) {
   var n = el.getElementsByTagName('BUTTON');
   if (n.length !== 1)
      throw new Error("BUG: Focus.PushButton id:" + el.id +
         " elements:" + n.length + ", expected: 1");
   return n[0];
});

eq.OnKey.set('PushButton', function(id, el, e) {
   var dm = eq.Dom;
   if (dm.isEnterOrSpace(e)) {
      dm.fireClick(e.target);
      eq.c.App.debounceKey(e.key);
      return true; // Consume event.
   }
   return false;
});

/*------------------------------------------------------------------------
   RadioButton callbacks.
------------------------------------------------------------------------*/

eq.Focus.set('RadioButton', function(el) {
   var n = el.getElementsByTagName('INPUT');
   if (n.length !== 1)
      throw new Error("BUG: Focus.RadioButton id:" + el.id +
         " elements:" + n.length + ", expected: 1");
   return n[0];
});

eq.OnKey.set('RadioButton', function(id, el, e) {
   // Click?
   var dm = eq.Dom;
   if (dm.isEnterOrSpace(e)) {
      dm.fireClick(e.target);
      eq.c.App.debounceKey(e.key);
      return true; // Consume event.
   }

   // Navigation?
   switch (e.key) {
      case 'Home':
      case 'PageUp':
      case 'End':
      case 'PageDown':
      case 'ArrowUp':
      case 'Up':
      case 'ArrowLeft':
      case 'Left':
      case 'ArrowDown':
      case 'Down':
      case 'ArrowRight':
      case 'Right':
         break;
      default:
         // No.
         return false;
   }

   var
      t = e.target,
      n = el.parentNode.querySelectorAll('input[name="' + t.name + '"]'),
      d = 0;
   if (n.length) {
      switch (e.key) {
         case 'Home':
         case 'PageUp':
            el = n[0];
            break;
         case 'End':
         case 'PageDown':
            el = n[v.length - 1];
            break;
         case 'ArrowUp':
         case 'Up':
         case 'ArrowLeft':
         case 'Left':
            d = -1;
            break;
         case 'ArrowDown':
         case 'Down':
         case 'ArrowRight':
         case 'Right':
            d = 1;
            break;
      }
      if (d !== 0) {
         el = null;
         for (var i = 0, l = n.length; i < l; i++)
            if (t === n[i]) {
               if ((i += d) === -1)
                  i = l - 1;
               else if (i >= l)
                  i = 0;
               el = n[i];
               break;
            }
      }
      if (el !== null)
         dm.fireClick(el);
   }

   eq.app.lastKey = e.key;
   return true; // Consume event.
});

eq.Changed.set('RadioButton', function(id, el) {
   if (el.length !== 1)
      throw new Error("BUG: Changed.RadioButton id:" + id +
         " elements:" + el.length + ", expected: 1");
   if (!el[0].checked)
      return null;
   var ch = {
      id : id,
      ty : eq.RsValType.buttonChecked,
      iv : 1
   };
   return ch;
});

/** @preserve $Id: 99-main.js,v 28.26 2021/03/17 07:43:55 mike Exp $
*//*======================================================================
   class App extends Obj
   Main thread application class.
========================================================================*/

eq.c.Obj.subClass('App', function() {
   eq.c.Obj.call(this);

/* NOTYET

   // DLG anchor elements.
   // If not empty, dialog is anchored on first element.
   // If first element is null, dialog is anchored on body element.
   // Additional anchors are used to anchor dialog children by 'cssid'.
   this.dlgAnchor = [];
*/

   // Dialog map.
   this.dlg = new Map();

   // Current dialog.
   this.currDlg = null;

   // Focus.
   this.focus = null;

   // Changed elements.
   this.changed = new Map();

   // Application is idle (not waiting for user input).
   this.idle = true;

   // Layout grid.
   this.grid = {
      // w : Width ch factor
      // h : Height em factor
   };

   // Obtain application URI from script query string 'app' parameter.
   var
      uri, app, host, port,
      re = new RegExp('^(http[s]?://)([^/]+)(/[^?#]+/)[^?#]*[?#]?.*$', 'i');
   try {
      // Obtain script base URI from script URI.
      uri = re.exec(document.querySelector('script#eq-webdlg').src);
      this.scriptBase = uri[1] + uri[2] + uri[3];
   }
   catch (e) {
      throw new Error("WEBDLG: <script id=\"eq-webdlg\"> not found");
   }

   try {
      // Optional query string app= parameter.
      this.app = /[?&]app=([^&]+)/i.exec(uri[0])[1];
   } catch (e) {}

   try {
      // Optional query string host= parameter.
      host = /[?&]host=([^&]+)/i.exec(uri[0])[1];
   } catch (e) {}

   try {
      // Optional query string port= parameter.
      port = /[?&]port=([^&]+)/i.exec(uri[0])[1];
   } catch (e) {}

   // Compose base and WebSocket URIs.
   uri = re.exec(location.href);
   host = host ? host : uri[2];
   if (port) {
      try {
         host = /^(.+:)[^:]*$/.exec(host)[1] + port;
      }
      catch (e) {
         host += ':' + port;
      }
   }
   this.appBase = uri[1] + host + '/';
   this.appWS = 'ws' + uri[1].substring(4).toLowerCase() + host + '/';
});

/*------------------------------------------------------------------------
   static App.start()
   Start application.
------------------------------------------------------------------------*/

eq.c.App.start = function(e) {
   var self = eq.app, appOverride;

   if (e !== undefined) {
      // Deactivate 'eq-start' elements.
      var el = document.body.getElementsByClassName('eq-start');
      for (var i = 0, l = el.length; i < l; i++) {
         // Check for 'data-app' application override.
         if (e.target === el[i] && ('app' in el[i].dataset))
            appOverride = el[i].dataset.app;

         // Remove 'click' handler.
         el[i].removeEventListener('click', eq.c.App.start);

         // Make disabled.
         el[i].classList.add('eq-disabled');
         if (el[i].disabled !== undefined)
            el[i].disabled = true;
      }
   }

/* NOTYET

   // Find 'eq-anchor' elements if any.
   var
      el = document.body.getElementsByClassName('eq-anchor'),
      l = el.length;
   if (l) {
      // First, find 'eq-dialog' anchor if present.
      for (var i = 0, iDialog; i < l; i++)
         if (el[i].id === 'eq-dialog') {
            iDialog = i;
            break;
         }

      // Set first anchor.
      self.dlgAnchor.push(iDialog !== undefined ? 'eq-dialog' : null);

      // Next, find additional anchors if any.
      for (var i = 0; i < l; i++)
         if (i !== iDialog && el[i].id)
            self.dlgAnchor.push(el[i].id);
   }
*/

   // Notify API thread to start application.

   var app;
   if (appOverride !== undefined)
      app = appOverride;
   else {
      if (self.app === undefined)
         throw new Error("Cannot start, application name undefined");
      app = self.app;
   }

   self.api.postMessage({
      o   : eq.ApiOp.openWebSocket,
      app : app,
      url : self.appWS
   });
};

/*------------------------------------------------------------------------
   static App.prepareStart()
   return: true if 'click' handlers are installed on 'eq-start' elements
           false if no 'eq-start' elements are present
------------------------------------------------------------------------*/

eq.c.App.prepareStart = function() {
   // Find 'eq-start' elements if any.
   var
      el = document.body.getElementsByClassName('eq-start'),
      l = el.length;
   if (!l) {
      // No 'eq-start' elements.
      return false;
   }

   // Setup 'eq-start' click handlers.
   for (var i = 0, l = el.length; i < l; i++) {
      el[i].addEventListener('click', this.start);
      el[i].classList.remove('eq-disabled');
      if (el[i].disabled !== undefined)
         el[i].disabled = false;
   }

   return true;
};

/*------------------------------------------------------------------------
   static App.stopped()
   Application stopped.
------------------------------------------------------------------------*/

eq.c.App.stopped = function(e) {
   var self = eq.app;

   // Close POPUP BOX.
   if (self.popupBox !== undefined) {
      self.popupBox.close();
      delete self.popupBox;
   }

   // Clear dialog map and current dialog.
   self.dlg.forEach(eq.c.App.cbCloseDlg);
   self.dlg.clear();
   self.currDlg = null;

   // Clear focus.
   self.focus = null;

   // Clear changed elements.
   self.changed.clear();

/* NOTYET

   // Clear 'eq-anchor' elements.
   self.dlgAnchor.length = 0;
*/

   // Clear pending 'mousedown' and 'click' events.
   delete self.pendingMouseDown;
   delete self.pendingClick;

   // Application is idle.
   self.idle = true;
};

eq.c.App.cbCloseDlg = function(d) {
   d.close();
};

/*------------------------------------------------------------------------
   static App.domReady()
   Continue main thread initialization when DOM is ready.
------------------------------------------------------------------------*/

eq.c.App.domReady = function(e) {
   var app = eq.c.App;
   // Notify API thread when Document is complete.
   if (document.readyState === 'complete')
      app.documentComplete();
   else
      addEventListener('load', app.documentComplete);

   // Prepare starting application.
   if (!app.prepareStart()) {
      // Immediately start application.
      app.start();
   }

   // This method/listener is now obsolete.
   if (e !== undefined)
      removeEventListener('DOMContentLoaded', app.domReady);
   delete app.domReady;
};

/*------------------------------------------------------------------------
   static App.documentComplete()
   Notify API thread when document is complete.
------------------------------------------------------------------------*/

eq.c.App.documentComplete = function(e) {
   var app = eq.c.App, self = eq.app;

   // Calculate grid.
   var et = document.createElement('P');
   et.style.position = 'absolute';
   et.style.visibility = 'hidden';
   et.style.zIndex = eq.Dom.zIndex(document.body) - 1;
   et.style.top = et.style.left = '0';

   // Reference width:1ch height:1em.
   var el = et.cloneNode();
   el.style.width = '1ch';
   el.style.height = '1em';
   document.body.appendChild(el);
   var rc = el.getBoundingClientRect();
   self.grid.w = rc.width;
   self.grid.h = rc.height;
   document.body.removeChild(el);

   // Letter 'X' width/height.
   el = et.cloneNode();
   el.textContent = "X";
   document.body.appendChild(el);
   rc = el.getBoundingClientRect();
   self.grid.w = rc.width / self.grid.w;
   self.grid.h = rc.height / self.grid.h;
   document.body.removeChild(el);

   // Notify API thread.
   self.api.postMessage({
      o : eq.ApiOp.documentComplete,
      d : self.grid
   });

   // This method/listener is now obsolete.
   if (e !== undefined)
      removeEventListener('load', app.documentComplete);
   delete app.documentComplete;
};

/*------------------------------------------------------------------------
   static App.apiReceived()
   Process message from API thread.
------------------------------------------------------------------------*/

eq.c.App.apiReceived = function(m) {
   var app = eq.c.App, self = eq.app, op = eq.MainOp, resume = false;
   for (var di = 0, dl = m.data.length; di < dl; di++) {
      var o = m.data[di].o, d = m.data[di].d;
      switch (o) {
         case op.applicationStopped:
            // Application has stopped.
            app.stopped();
            if (d) {
               // Prepare restarting if necessary.
               app.prepareStart();
            }
            continue;

         case op.applicationFailed:
            // Prepare restarting if necessary.
            app.stopped();
            app.prepareStart();
            continue;

         case op.cancelCurrentCall:
            if (self.currDlg !== null) {
               self.setIdle(self.currDlg);
               self.currDlg = null;
               self.focus = null;
               self.changed.clear();
            }
            else if (self.popupBox !== undefined) {
               self.popupBox.close();
               delete self.popupBox;
            }
            continue;
      }

      if (!self.idle)
         throw new Error("apiReceived op + " + o + ": invalid, not idle");
      if (self.currDlg !== null)
         throw new Error("apiReceived op + " + o +
            ": invalid, current dialog pending");
      if (self.focus !== null)
         throw new Error("apiReceived op + " + o + ": invalid, focus pending");
      if (self.changed.size)
         throw new Error("apiReceived op + " + o + ": invalid, " +
            size + " changed elements pending");
      if (self.popupBox !== undefined)
         throw new Error("apiReceived op + " + o + ": invalid, POPUP BOX open");

      switch (o) {
         case op.resume:
            resume = true;
            break;

         case op.addControl: {
            var
               id = m.data[di].i,
               dlg = self.dlg.get(id);
            if (d.pa !== undefined) {
               // Add to specified parent.
               if (dlg === undefined)
                  throw new Error("apiReceived addControl: DLG " +
                     id + " does not exist");
               var
                  p = document.getElementById(d.pa),
                  el;
               if (!p)
                  throw new Error("apiReceived addControl: parent '" +
                     d.pa + "' does not exist");
               p = p.getElementsByClassName('eq-container');
               if (p.length !== 1)
                  throw new Error("apiReceived addControl: parent '" +
                     d.pa + "' invalid containers: " +
                     p.length + ", expected: 1");
               el = dlg.addControl(d, p[0]);
/* NOTYET
               if (self.dlgAnchor.indexOf(d.pa) !== -1) {
                  el.classList.add('eq-root');
                  dlg.root.push(el);
                  app.addRootListeners(el);
                  self.setIdle(dlg);
               }
*/
            }
            else {
               // Add to root.
               if (dlg === undefined)
                  dlg = self.newDialog(id);
               // TODO: Multiple dialogs
               dlg.close();
               var
                  el = dlg.addControl(d),
                  p;
               el.style.display = 'none';
               el.classList.add('eq-root');
               // TODO: Multiple dialogs, dlgAnchor
               // For now, attach to 'eq-dialog' element or body.
               if (!(p = document.getElementById('eq-dialog')))
                  p = document.body;
               p.appendChild(el);
               dlg.root.push(el);
               app.addRootListeners(el);
               self.setIdle(dlg);
            }
            break;
         }

         case op.modifyControl: {
            var
               id = m.data[di].i,
               dlg = self.dlg.get(id);
            if (dlg === undefined)
               throw new Error("apiReceived modifyControl: DLG " +
                  id + " does not exist");
            if (!d.id)
               throw new Error("apiReceived modifyControl: id not set");
            if (!(el = document.getElementById(d.id)))
               throw new Error("apiReceived modifyControl: id '" +
                  d.id + "' does not exist");
            dlg.modifyControl(d.id, el, d);
            break;
         }

         case op.deleteControl: {
            var
               id = m.data[di].i,
               dlg = self.dlg.get(id),
               el = document.getElementById(d),
               i;
            if (dlg === undefined)
               throw new Error("apiReceived deleteControl: DLG " +
                  id + " does not exist");
            if (!el)
               throw new Error("apiReceived deleteControl: id '" +
                  d + "' does not exist");
            // Root element?
            if ((i = dlg.root.indexOf(el)) !== -1) {
               // Remove from DLG root elements, remove root listeners.
               dlg.root.splice(i, 1);
               if (dlg.root.length === 0)
                  self.closeDialog(dlg);
               // TODO: dialog deleted, root children no longer present in DOM
               else
                  app.removeRootListeners(el);
            }
            else {
               // Remove from element descriptors.
               dlg.eds.delete(d);
            }
            // Remove from DOM.
            el.parentNode.removeChild(el);
            break;
         }

         case op.invalidate: {
            for (var i = d.length; --i >= 0;) {
               var dlg = self.dlg.get(d[i]);
               if (dlg !== undefined)
                  self.closeDialog(dlg);
            }
            break;
         }

         case op.dlg_DRAW: {
            if (resume) {
               resume = false;
               self.dialogsVisible();
            }
            var
               id = m.data[di].i,
               dlg = self.dlg.get(id);
            if (dlg === undefined)
               throw new Error("apiReceived DLG DRAW: DLG " +
                  id + " does not exist");
            // Make Dialog top and visible, immediately reset currDlg.
            self.setCurrDialog(dlg);
            self.currDlg = null;
            break;
         }

         case op.dlg_DO: {
            if (resume) {
               resume = false;
               self.dialogsVisible();
            }
            var
               id = m.data[di].i,
               dlg = self.dlg.get(id);
            if (dlg === undefined)
               throw new Error("apiReceived DLG DO: DLG " +
                  id + " does not exist");
            // Make Dialog top and visible.
            self.setCurrDialog(dlg);
            if (d.f)
               self.setFocus(d.f);
            if (d.t !== undefined)
               dlg.tabOrder = d.t;
            // Activate dialog.
            self.setIdle();
            break;
         }

         case op.dlg_POPUP_BOX: {
            if (resume) {
               resume = false;
               self.dialogsVisible();
            }
            var p = self.topDialog();
            if (p && p.root.length) {
               if (!(p = p.root[0].querySelector('.eq-root>.eq-idle')))
                  throw new Error("apiReceived POPUP BOX: top dialog not idle");
            }
            else
               p = app.createIdlePane(document.body);
            p.classList.add('eq-fog');
            self.popupBox = new eq.c.PopupBox(d, p);
            break;
         }

         default:
            throw new Error("apiReceived invalid op: " + o);
      }
   }

   if (resume)
      self.dialogsVisible();
};

/*------------------------------------------------------------------------
   static App.onEvent()
   Root element event handler.
------------------------------------------------------------------------*/

eq.c.App.onEvent = function(e) {
   var self = eq.app;
   var md = e.type === 'mousedown';
   if (md) {
      // 'mousedown', invalidate last key.
      delete self.lastKey;
   }

   if (self.idle) {
      if (e.target.classList.contains('eq-idle')) {
         // Event on idle pane, set wait cursor.
         e.target.style.cursor = 'wait';
         eq.Dom.consumeEvent(e);
      }
      return;
   }

   var
      el = eq.Dom.eqControl(e.target), // Obtain 'eq-control' element.
      mdFocus = false;
   if (el) {
      if (el.classList.contains('eq-disabled')) {
         eq.Dom.consumeEvent(e);
         return;
      }
      var id = el.id, dlg = self.currDlg, ed;
      if (!dlg)
         throw new Error("BUG: onEvent no current dialog");
      if ((ed = dlg.eds.get(id)) !== undefined) {

         // 'mousedown'?
         if (md) {
            mdFocus = el.classList.contains('eq-focus');
            if (self.pendingMouseDown !== undefined)
               throw new Error("BUG: onEvent pending mousedown id:" +
                  self.pendingMouseDown.id);
            self.pendingMouseDown = { id : id, t : e.target };
         }
         else if (e.type === 'focusin') {
            // About to receive focus.
            if (el.classList.contains('eq-focus')) {
               // Set focus if different from current focus.
               if (id !== self.focus)
                  self.setFocus(id, el);
            }
            else {
               // Target not focusable, ignore event.
               eq.Dom.consumeEvent(e);
            }
            return;
         }

         self.processEvent(e, id, el, ed);
      }
   }

   if (md && !mdFocus) {
      // 'mousedown', target not focusable.
      eq.Dom.consumeEvent(e);
   }
};

/*------------------------------------------------------------------------
   App.processEvent()
   Process change and action event.
------------------------------------------------------------------------*/

eq.c.App.prototype.processEvent = function(e, id, el, ed) {
   // Change event?
   if (ed.c !== undefined && ed.c.indexOf(e.type) !== -1) {
      // Yes, enqueue change if not already in queue.
      var ch = this.changed.get(id), cb;
      if (ch === undefined)
         this.changed.set(id, [ e.type ]);
      else if (ch.indexOf(e.type) === -1)
         ch.push(e.type);
      // Invoke 'OnChange' handler if present.
      if ((cb = eq.OnChange.get(ed.cb)) !== undefined)
         cb(id, el, e);
   }

   // Action event?
   if (ed.a !== undefined && ed.a.indexOf(e.type) !== -1) {
      if (e.type === 'mousedown') {
         // Yes, 'mousedown', register pending 'click'.
         if (this.pendingClick !== undefined)
            throw new Error("BUG: onEvent pending click id:" +
               this.pendingClick.id + " type:" +
               this.pendingClick.type);
         this.pendingClick = {
            id : id,
            type : 'click'
         };
      }
      else if (this.pendingClick !== undefined) {
         // Yes, 'click' pending, append event.
         if (this.pendingClick.ev !== undefined)
            throw new Error("BUG: onEvent pending click id:" +
               this.pendingClick.id + " type:" +
               this.pendingClick.type + " and id:" +
               this.pendingClick.ev.id + " type:" +
               this.pendingClick.ev.type);
         this.pendingClick.ev = {
            id : id,
            type : e.type
         };
      }
      else {
         // Yes, submit event.
         this.submit(el, {
            id : id,
            type : e.type
         });
      }

      return true;
   }

   return false;
};

/*------------------------------------------------------------------------
   App.setChanged()
   Set element changed.
------------------------------------------------------------------------*/

eq.c.App.prototype.setChanged = function(id, ty) {
   var dlg = this.currDlg, ed;
   if (!dlg)
      throw new Error("BUG: onEvent no current dialog");
   if ((ed = dlg.eds.get(id)) !== undefined) {
      if (ed.c !== undefined && ed.c.indexOf(ty) !== -1) {
         var ch = this.changed.get(id);
         if (ch === undefined)
            this.changed.set(id, [ ty ]);
         else if (ch.indexOf(ty) === -1)
            ch.push(ty);
      }
   }
};

/*------------------------------------------------------------------------
   App.elementsChanged()
   Collect and return element changes.
------------------------------------------------------------------------*/

eq.c.App.prototype.elementsChanged = function() {
   var changed = [{
      ty : eq.RsValType.idFocus,
      sv : this.focus !== null ? this.focus : ""
   }];
   this.changed.forEach(eq.c.App.cbElementsChanged, { c : changed });
   this.changed.clear();
   return changed;
};

eq.c.App.cbElementsChanged = function(ty, id) {
   var
      dlg = eq.app.currDlg,
      ed, el, cb, v;
   if (!dlg)
      throw new Error("BUG: elementsChanged no current dialog");
   if ((ed = dlg.eds.get(id)) === undefined)
      throw new Error("BUG: elementsChanged id:" + id +
         " not registered, type:" + ty);
   if ((cb = eq.Changed.get(ed.cb)) === undefined)
      throw new Error("BUG: elementsChanged id:" + id +
         " 'Changed' handler undefined, type:" + ty);
   el = document.getElementById(id).getElementsByClassName('eq-event');
   if ((v = cb(id, el, ty)) !== null)
      this.c.push(v);
};

/*------------------------------------------------------------------------
   App.submit()
   Submit event.
------------------------------------------------------------------------*/

eq.c.App.prototype.submit = function(el, ev) {
   // Special case for CheckBox, RadioButton,
   // enqueue change if not already in queue.
   if (   (   el.classList.contains('eq-checkbox')
           || el.classList.contains('eq-radiobutton'))
       && !this.changed.has(el.id))
      this.changed.set(el.id, 'input');
   // Notify API thread on next event cycle.
   var app = eq.c.App;
   this.setIdle(this.currDlg);
   app.debounceLastKey();
   eq.Dom.onNextCycle(app.apiSubmit, ev);
}

eq.c.App.apiSubmit = function(ev) {
   var self = eq.app;
   ev.changed = self.elementsChanged();
   self.api.postMessage({
      o : eq.ApiOp.onEvent,
      d : ev
   });
   // Clear current dialog and focus.
   self.currDlg = null;
   self.focus = null;
};

/*------------------------------------------------------------------------
   static App.onKeyDown() .onKeyUp()
   Global keyboard event handler.
------------------------------------------------------------------------*/

eq.c.App.onKeyDown = function(e) {
   var self = eq.app;
   if (e.key === self.debounceKey) {
      eq.c.App.debounceKey(e.key);
      eq.Dom.consumeEvent(e);
      return;
   }

   if (self.popupBox !== undefined) {
      // POPUP BOX is open.
      if (self.popupBox.onKey(e))
         eq.Dom.consumeEvent(e);
      return;
   }

   // Obtain target element, 'eq-control' or focus.
   var el = eq.Dom.eqControl(e.target);
   if (!el && e.target === document.body && self.focus !== null)
      el = document.getElementById(self.focus);
   if (el) {
      if (el.classList.contains('eq-disabled')) {
         eq.Dom.consumeEvent(e);
         return;
      }
      var dlg = self.currDlg, id, ed;
      if (dlg && (ed = dlg.eds.get(id = el.id)) !== undefined) {
         // Target is DLG element.
         if (self.idle) {
            // Ignore event while idle.
            // TODO: type-ahead
            eq.Dom.consumeEvent(e);
            return;
         }

         // Navigation or function key?
         // TODO: function keys
         switch (e.key) {
            case 'Enter':
               if (el.classList.contains('eq-noenter'))
                  break;
               // FALLTHROUGH
            case 'Tab': {
               // Navigate to next/previous element in tab order.
               var tb = dlg.tabOrder;
               if (tb !== null) {
                  var i = tb.indexOf(id);
                  if (i !== -1) {
                     var
                        shift = e.shiftKey,
                        j = shift ? i - 1 : i + 1,
                        l = tb.length;
                     while (true) {
                        if (shift) {
                           if (j === -1)
                              j = l - 1;
                        }
                        else {
                           if (j === l)
                              j = 0;
                        }
                        if (j === i)
                           break;

                        el = document.getElementById(tb[j]);
                        if (el && !el.classList.contains('eq-notab')) {
                           // Found next/previous element in tab order.
                           self.setFocus(tb[j], el);
                           break;
                        }

                        if (shift)
                           j--;
                        else
                           j++;
                     }
                  }
               }
               self.lastKey = e.key;
               eq.Dom.consumeEvent(e);
               return;
            }
         }

         var cb = eq.OnKey.get(ed.cb);
         if (cb !== undefined && cb(id, el, e))
            eq.Dom.consumeEvent(e);
      }
   }
};

eq.c.App.onKeyUp = function(e) {
   var self = eq.app;
   if (e.key === self.debounceKey) {
      clearTimeout(self.debounceKeyTimer);
      delete self.debounceKeyTimer;
      delete self.debounceKey;
      eq.Dom.consumeEvent(e);
   }
};

eq.c.App.debounceLastKey = function() {
   var key = eq.app.lastKey;
   if (key !== undefined)
      this.debounceKey(key);
};

eq.c.App.debounceKey = function(key) {
   var self = eq.app;
   self.debounceKey = key;
   if (self.debounceKeyTimer !== undefined)
      clearTimeout(self.debounceKeyTimer);
   self.debounceKeyTimer = setTimeout(this.onDebounceKeyTimer, 1000);
   delete self.lastKey;
};

eq.c.App.onDebounceKeyTimer = function() {
   var self = eq.app;
   // Timer has elapsed.
   delete self.debounceKeyTimer;
   delete self.debounceKey;
};

/*------------------------------------------------------------------------
   static App.onMouseUp()
   Global 'mouseup' event handler.
------------------------------------------------------------------------*/

eq.c.App.onMouseUp = function(e) {
   var self = eq.app;
   if (self.idle) {
      // 'mouseup' while idle.
      delete self.pendingMouseDown;
      // 'click' must not be pending.
      if (self.pendingClick !== undefined) {
         var id = self.pendingClick.id;
         delete self.pendingClick;
         throw new Error("BUG: idle onMouseUp pending click id:" + id);
      }
      return;
   }

   if (self.pendingMouseDown !== undefined) {
      // 'mouseup' while 'mousedown' pending.
      var t = self.pendingMouseDown.t;
      if (t === e.target) {
         // 'mouseup' target same as pending 'mousedown' target,
         // obtain 'eq-control' element.
         var
            id = self.pendingMouseDown.id,
            dlg = self.currDlg,
            el = eq.Dom.eqControl(t);
         if (!dlg)
            throw new Error("BUG: onMouseUp no current dialog");
         // 'click' committed.
         self.processEvent({
            type : 'click',
            target : t }, id, el, dlg.eds.get(id));
         if (el.classList.contains('eq-focus')) {
            // Element focusable, set focus.
            self.setFocus(id, el);
         }
      }

      delete self.pendingMouseDown;
   }

   if (self.pendingClick !== undefined) {
      // 'mouseup' while 'click' pending,
      // obtain 'eq-control' element.
      var el = eq.Dom.eqControl(e.target);
      if (el && el.id === self.pendingClick.id) {
         // 'mouseup' target same as pending 'click' target,
         // 'click' committed, submit event.
         self.submit(el, self.pendingClick);
      }

      delete self.pendingClick;
   }
}

/*------------------------------------------------------------------------
   static App.onPopupClick()
   Global POPUP BOX 'click' Event handler.
------------------------------------------------------------------------*/

eq.c.App.onPopupClick = function(e) {
   var self = eq.app;
   if (self.idle && self.popupBox !== undefined) {
      if (self.popupBox.isEvent(e)) {
         if (self.popupBox.close(e)) {
            delete self.popupBox;
            return;
         }
      }
      else if (!e.target.classList.contains('eq-idle'))
         return;

      if (e.type === 'mousedown') {
         // 'mousedown', target not focusable.
         e.preventDefault();
      }
   }
}

/*------------------------------------------------------------------------
   static App.addRootListeners() .removeRootListeners()
   Add/remove root element event listeners.
------------------------------------------------------------------------*/

eq.c.App.rootEventTypes = [
   'focusin',
   'mousedown',
   'input'
];

eq.c.App.addRootListeners = function(el) {
   var ln = this.onEvent, types = this.rootEventTypes;
   for (var i = 0, l = types.length; i < l; i++)
      el.addEventListener(types[i], ln, true);
}

eq.c.App.removeRootListeners = function(el) {
   var ln = this.onEvent, types = this.rootEventTypes;
   for (var i = 0, l = types.length; i < l; i++)
      el.removeEventListener(types[i], ln, true);
}

/*------------------------------------------------------------------------
   App.setFocus()
   Set focus by element id.
------------------------------------------------------------------------*/

eq.c.App.prototype.setFocus = function(id, el) {
   if (id !== this.focus) {
      var
         dlg = this.currDlg,
         n, f, ed, cb;
      if (!dlg)
         throw new Error("BUG: setFocus no current dialog");

      // Remove current focus indicator.
      n = document.body.getElementsByClassName('eq-focused');
      for (var i = 0, l = n.length; i < l; i++)
         n[i].classList.remove('eq-focused');

      if ((f = this.focus) !== null) {
         // Focus change, check 'blur' event on current focus element.
         if ((ed = dlg.eds.get(f)) === undefined)
            throw new Error("setFocus: from '" + f + "' not registered");
         var e, from = document.getElementById(f);
         if (!from)
            throw new Error("setFocus: from '" + f + "' does not exist");
         cb = eq.Focus.get(ed.cb);
         e = {
            type : 'blur',
            target : cb !== undefined ? cb(from) : from
         };
         if (this.processEvent(e, f, from, ed)) {
            // Action event submitted, done.
            this.focus = id;
            return;
         }
      }

      // Set new focus.
      if ((ed = dlg.eds.get(id)) === undefined)
         throw new Error("setFocus: to '" + id + "' not registered");
      if (el === undefined && !(el = document.getElementById(id)))
         throw new Error("setFocus: to '" + id + "' does not exist");
      if ((cb = eq.Focus.get(ed.cb)) !== undefined)
         el = cb(el);

      this.focus = id;
      el.classList.add('eq-focused');
      el.focus();
      // Verify focus on next event cycle.
      eq.Dom.onNextCycle(eq.c.App.verifyFocus);
   }
};

eq.c.App.verifyFocus = function() {
   var f = eq.app.focus, el;
   if (f && (!(el = eq.Dom.eqControl(document.activeElement)) || el.id !== f))
      // Focus could not be set, set focus to body element.
      if ((el = document.activeElement) && el !== document.body)
         el.blur();
};

/*------------------------------------------------------------------------
   App.setIdle()
   Set application idle or active.
------------------------------------------------------------------------*/

eq.c.App.prototype.setIdle = function(dlg) {
   var app = eq.c.App;
   this.stopIdleTimer();
   if (dlg !== undefined) {
      // Idle.
      this.idle = true;
      // Add idle pane(s) if necessary.
      var added = false;
      for (var i = 0, l = dlg.root.length; i < l; i++) {
         var el = dlg.root[i];
         if (!el.querySelector('.eq-root>.eq-idle')) {
            app.createIdlePane(el);
            added = true;
         }
         // After 5 seconds, visualize idle state on newly added panes.
         if (added)
            this.idleTimer = setTimeout(app.onIdleTimer, 5000);
      }
   }
   // Delay setting active state until next event cycle.
   else
      setTimeout(app.setActive);
};

// Set active state.
eq.c.App.setActive = function() {
   var
      self = eq.app,
      dlg = self.currDlg;
   if (!dlg)
      throw new Error("BUG: setActive no current dialog");
   self.idle = false;
   // Remove idle pane(s) if any.
   for (var i = 0, l = dlg.root.length; i < l; i++) {
      var el = dlg.root[i].querySelector('.eq-root>.eq-idle');
      if (el)
         el.parentNode.removeChild(el);
   }
};

// Visualze idle state.
eq.c.App.onIdleTimer = function() {
   // Timer has elapsed.
   delete eq.app.idleTimer;
   // Add 'eq-fog' class to idle pane(s) if necessary.
   var el = document.body.getElementsByClassName('eq-idle');
   for (var i = 0, l = el.length; i < l; i++)
      el[i].classList.add('eq-fog');
};

/*------------------------------------------------------------------------
   static App.createIdlePane()
   Create idle pane.
------------------------------------------------------------------------*/

eq.c.App.createIdlePane = function(parent) {
   var el = document.createElement('DIV');
   el.className = 'eq-idle';
   el.style.zIndex = eq.Dom.zIndex(parent) + 1;
   parent.appendChild(el);
   return el;
};

/*------------------------------------------------------------------------
   App.stopIdleTimer()
   Stop application idle timer.
------------------------------------------------------------------------*/

eq.c.App.prototype.stopIdleTimer = function() {
   if (this.idleTimer !== undefined) {
      clearTimeout(this.idleTimer);
      delete this.idleTimer;
   }
};

/*------------------------------------------------------------------------
   App.newDialog()
   Create new dialog.
------------------------------------------------------------------------*/

eq.c.App.prototype.newDialog = function(id) {
   // TODO: Multiple dialogs
   // For now, invalidate and close existing dialogs.
   this.dlg.forEach(eq.c.App.cbInvalidateCloseDlg, this);
   this.dlg.clear();
   this.currDlg = null;

   var dlg = new eq.c.Dlg(id);
   this.dlg.set(id, dlg);
   return dlg;
};

eq.c.App.cbInvalidateCloseDlg = function(d) {
   this.api.postMessage({
      o : eq.ApiOp.invalidate,
      d : d.id
   });
   d.close();
};

/*------------------------------------------------------------------------
   App.closeDialog()
   Close dialog.
------------------------------------------------------------------------*/

eq.c.App.prototype.closeDialog = function(dlg) {
   if (dlg === this.currDlg)
      this.currDlg = null;
   this.dlg.delete(dlg.id);
   dlg.close();
};

/*------------------------------------------------------------------------
   App.topDialog()
   Get topmost dialog.
------------------------------------------------------------------------*/

eq.c.App.prototype.topDialog = function() {
   var v = {};
   this.dlg.forEach(eq.c.App.cbTopDialog, v);
   return v.d !== undefined ? v.d : null;
};

eq.c.App.cbTopDialog = function(d) {
   this.d = d;
};

/*------------------------------------------------------------------------
   App.setCurrDialog()
   Set current dialog, make it top and visible.
------------------------------------------------------------------------*/

eq.c.App.prototype.setCurrDialog = function(dlg) {
   if (dlg !== this.topDialog()) {
      var id = dlg.id;
      this.dlg.delete(id);
      this.dlg.set(id, dlg);
   }

   dlg.visible();
   this.currDlg = dlg;
};

/*------------------------------------------------------------------------
   App.dialogsVisible()
   Make dialogs visible.
------------------------------------------------------------------------*/

eq.c.App.prototype.dialogsVisible = function() {
   this.dlg.forEach(eq.c.App.cbDialogsVisible);
};

eq.c.App.cbDialogsVisible = function(d) {
   d.visible();
};

/*========================================================================
   class Dlg extends Obj
   DLG dialog instance.
========================================================================*/

eq.c.Obj.subClass('Dlg', function(id) {
   eq.c.Obj.call(this);

   // Dialog id.
   this.id = id;

   // Root elements.
   this.root = [];

   // Element descriptors.
   this.eds = new Map();

   // Tab order.
   this.tabOrder = null;
});

/*------------------------------------------------------------------------
   Dlg.close()
   Close dialog instance.
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.close = function() {
   var app = eq.c.App;

   // Remove root elements.
   for (var i = this.root.length; --i >= 0;) {
      var el = this.root[i];
      app.removeRootListeners(el);
      el.parentNode.removeChild(el);
   }
   this.root.length = 0;

   // Clear element descriptors.
   this.eds.clear();

   // Clear tab order.
   this.tabOrder = null;
};

/*------------------------------------------------------------------------
   Dlg.visible()
   Make dialog visible.
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.visible = function() {
   for (var i = this.root.length; --i >= 0;) {
      var el = this.root[i];
      el.style.visibility = '';
      el.style.display = '';
   }
};

/*------------------------------------------------------------------------
   Dlg.addControl()
   Add DLG control to document from DOM element template(s).
   See: dlg.js Dlg.prototype.addChild()
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.addControl = function(et, elParent, rootId) {
   var root;
   if (rootId === undefined && !(rootId = et.id))
      throw new Error("addControl: root id not set");

   while (true) {
      // Create element.
      var el = document.createElement(et.tg !== undefined ? et.tg : 'DIV');
      this.updateControl(rootId, el, et);
      if (root === undefined)
         root = el;

      // Add to DOM.
      if (elParent !== undefined)
         elParent.appendChild(el);
      elParent = el;

      // Nested template.
      if (et.et === undefined)
         break;
      et = et.et;
   }

   // Children.
   if (et.ch !== undefined)
      for (var i = 0, l = et.ch.length; i < l; i++) {
         var id = rootId, el = et.ch[i];
         if (el.id)
            id = el.id;
         else
            el.id = id + '-' + (i+1);
         this.addControl(el, elParent, id);
      }

   return root;
};

/*------------------------------------------------------------------------
   Dlg.modifyControl()
   Modify DLG control from DOM element template(s).
   See: dlg.js Dlg.prototype.addChild()
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.modifyControl = function(rootId, el, et) {
   while (true) {
      if (et.tg !== undefined && et.tg !== el.tagName) {
         // Tag modified, recreate element.
         var elNew = document.createElement(et.tg);
         el.parentNode.replaceChild(elNew, el);
         el = elNew;
      }

      // Update element.
      this.updateControl(rootId, el, et);

      // Nested template.
      if (et.et === undefined)
         break;
      if (el.childNodes.length !== 1)
         throw new Error("modifyControl: id '" + rootId +
            "' nested template, childNodes: " + el.childNodes.length +
            ", expected: 1");
      et = et.et;
      el = el.childNodes[0];
   }

   // Children.
   if (et.ch !== undefined) {
      for (var i = 0, l = et.ch.length; i < l; i++) {
         if (i >= el.childNodes.length)
            el.appendChild(document.createElement(
               et.ch[i].tg !== undefined ? et.ch[i].tg : 'DIV'));
         this.modifyControl(rootId, el.childNodes[i], et.ch[i]);
      }
      while (el.childNodes.length > l)
         el.removeChild(el.childNodes[l]);
   }
};

/*------------------------------------------------------------------------
   Dlg.updateControl()
   Update DLG control from DOM element template.
   See: dlg.js Dlg.prototype.addChild()
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.updateControl = function(rootId, el, et) {
   for (var a in et)
      switch (a) {
         case 'id':
            // id
            el.id = et.id;
            break;

         case 'cb': {
            // Callback.
            var id = el.id, ed;
            if (!id)
               throw new Error("BUG: updateControl undefined callback id");
            if ((ed = this.eds.get(id)) !== undefined)
               ed.cb = et.cb;
            else
               this.eds.set(id, { cb : et.cb });
            break;
         }

         case 'tc':
            // Text content.
            el.textContent = et.tc;
            break;

         case 'hc':
            // Text or HTML content.
            eq.Dom.setContent(el, et.hc);
            break;

         case 'ec':
            // EditText content.
            el.value = et.ec;
            if (el.tagName === 'TEXTAREA')
               el.textContent = et.ec;
            break;

         case 'at': {
            // Attributes.
            for (var at in et.at) {
               var a = et.at[at];
               if (a !== null)
                  el.setAttribute(at, a);
               else
                  el.removeAttribute(at);
            }
            break;
         }

         case 'clr': {
            // Remove class names.
            for (var i = 0, l = et.clr.length; i < l; i++)
               el.classList.remove(et.clr[i]);
            break;
         }

         case 'cla': {
            // Add class names.
            for (var i = 0, l = et.cla.length; i < l; i++)
               el.classList.add(et.cla[i]);
            break;
         }

         case 'st': {
            // Styles.
            for (var st in et.st) {
               var s = et.st[st];
               el.style.setProperty(st, s !== null ? s : '');
            }
            break;
         }

         case 'rt': {
            // Value by request tag, invoke setAttr callback.
            var
               ed = this.eds.get(rootId),
               cb = eq.setAttr.get(ed.cb);
            if (cb === undefined)
               throw new Error("BUG: updateControl id:" + rootId +
                  " setAttr callback undefined");
            cb(el, et.rt);
            break;
         }

         case 'ev': {
            // Events.
            var
               ed = this.eds.get(rootId),
               ev = et.ev,
               enabled = false;
            if (ed === undefined)
               throw new Error("BUG: updateControl id:" + rootId +
                  " not registered");
            if (ev.a !== undefined) {
               if (ev.a !== null) {
                  ed.a = ev.a;
                  enabled = true;
               }
               else
                  delete ed.a;
            }
            if (ev.c !== undefined) {
               if (ev.c !== null) {
                  ed.c = ev.c;
                  enabled = true;
               }
               else
                  delete ed.c;
            }
            if (enabled)
               el.classList.add('eq-event');
            else
               el.classList.remove('eq-event');
            break;
         }
      }
};

/*========================================================================
   Start main thread
========================================================================*/

eq.app = new eq.c.App();

// Install global event listeners.
addEventListener('keydown', eq.c.App.onKeyDown, true);
addEventListener('keyup', eq.c.App.onKeyUp, true);
addEventListener('mouseup', eq.c.App.onMouseUp, true);
addEventListener('mousedown', eq.c.App.onPopupClick, true);
addEventListener('click', eq.c.App.onPopupClick, true);

// Start API thread.
eq.app.api = new Worker(eq.app.scriptBase + "eq-webdlg-api.js");
eq.app.api.onmessage = eq.c.App.apiReceived;
eq.app.api.onerror = function(e) {
   throw new Error("WEBDLG API thread failed in line " +
      e.lineno + ": " + e.message);
};

// Continue initialization when DOM is ready.
if (document.readyState === 'loading')
   addEventListener('DOMContentLoaded', eq.c.App.domReady);
else
   eq.c.App.domReady();

